/*
 *
 * Copyright (c) 1994
 * Hewlett-Packard Company
 *
 * Permission to use, copy, modify, distribute and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.  Hewlett-Packard Company makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 *
 */

/*
The Safe STL enhancements are

Copyright (C) Horstmann Software Design Corp. 1995. All Rights Reserved.

You may freely use, copy, modify, or distribute these Safe STL enhancements
and this documentation for any purpose, provided (1) you charge NO FEE for
distributing these copyrighted materials and (2) you include both the
copyright notice and permission notice from Hewlett-Packard Company and this
copyright notice and permission notice. Horstmann Software Design Corporation
specifically disclaims that these materials might do anything useful at all.
If they work for you and don't destroy anything, you are in luck.
There is no free technical support available for these materials.

I marked all modifications to the STL source with CSH (Cay S. Horstmann)

Please send bug reports/suggestions to horstman@cs.sjsu.edu.
For the most current version of these files, see
http://www.mathcs.sjsu.edu/faculty/horstman/safestl.html
*/

#ifndef DEQUE_H
#define DEQUE_H

#include <function.h>
// CSH #include <algobase.h>
#include <algorith.h> // CSH
#include <iterator.h>
// CSH #include <bool.h>
// CSH begin
#ifndef assert
#include <assert.h>
#endif
// CSH end

// CSH begin
namespace std {
// CSH end

#ifndef Allocator
#define Allocator allocator
// CSH #include <defalloc.h>
#include <memory.h> // CSH
#endif

#ifndef deque 
#define deque deque
#endif

template <class T>
class deque {
public:
	 class iterator;
    class const_iterator;
    friend class iterator;
    friend class const_iterator;
public:
    typedef T value_type;
    typedef Allocator<T> data_allocator_type;
    typedef Allocator<T>::pointer pointer;
	 typedef Allocator<T>::reference reference;
	 typedef Allocator<T>::const_reference const_reference;
	 typedef Allocator<T>::size_type size_type;
    typedef Allocator<T>::difference_type difference_type;
    typedef Allocator<pointer> map_allocator_type;
protected:
	 static data_allocator_type data_allocator;
	 static size_type buffer_size;
	 static map_allocator_type map_allocator;
	 typedef Allocator<pointer>::pointer map_pointer;
public:

// CSH begin
			 class iterator;
			 class const_iterator;

			 class iterator_base
			 {
          protected: // CSH
          	typedef deque<T> owner_type;
          private: // CSH
				friend class deque<T>;
				friend class iterator;
				friend class const_iterator;

				 iterator_base* _next_iter;
				 owner_type* _owner;
			 protected:
	pointer current;
	pointer first;
	pointer last;
	map_pointer node;
				 iterator_base() : _owner(0) {}
				 bool valid(difference_type n = 0) const
				 {  if (_owner == 0) return false;
					 if (node == 0) return false;

					 difference_type offset = n + (current - first);
					 if (offset >= 0)
					 {	difference_type node_to_jump = offset / buffer_size;
						return node + node_to_jump < _owner->finish_node
						|| node + node_to_jump == _owner->finish_node
						&& current < _owner->finish_current;
					 }
					 else
					 {	difference_type node_to_jump = -((-offset + buffer_size - 1) / buffer_size);

						return node - node_to_jump > _owner->start_node
						|| node - node_to_jump == _owner->start_node
						&& current >= _owner->start_current;
					 }
				 }
				 bool valid_or_past_end(difference_type n = 0) const
				 {  if (_owner == 0) return false;
					 if (node == 0) return false;

					 difference_type offset = n + (current - first);
					 if (offset >= 0)
					 {	difference_type node_to_jump = offset / buffer_size;
						return node + node_to_jump < _owner->finish_node
						|| node + node_to_jump == _owner->finish_node
						&& current <= _owner->finish_current;
					 }
					 else
					 {	difference_type node_to_jump = -((-offset + buffer_size - 1) / buffer_size);

						return node - node_to_jump > _owner->start_node
						|| node - node_to_jump == _owner->start_node
						&& current >= _owner->start_current;
					 }
				 }
				 bool orphan() const
				 {  return _owner == 0;
				 }
				 bool same_owner(const iterator_base& b) const
				 {  if (orphan() || b.orphan()) return false;
					 return _owner == b._owner;
				 }
				 void attach(pointer c, map_pointer n, owner_type* owner)
				 {  current = c; node = n;  _owner = owner;
					 if (n != 0) { first = *n, last = *n + buffer_size; }
					 else { first = 0; last = 0; }
					 // add into iterator chain of owner
					 _next_iter = _owner->_iter_chain;
					 _owner->_iter_chain = this;
				 }

				 void detach()
				 {  // remove from iterator chain of owner
					 if (_owner)
					 {  iterator_base* p = _owner->_iter_chain;
						 iterator_base* ppre = 0;
						 while (p != 0 && p != this)
						 {  ppre = p;
							 p = p->_next_iter;
						 }
						 if (ppre != 0)
							 ppre->_next_iter = p->_next_iter;
						 else
							 _owner->_iter_chain = p->_next_iter;
						 _owner = 0;
					 }
				 }

		 };
		 friend class iterator_base;
// CSH end

	 class iterator : public random_access_iterator<T, difference_type>,
	 	public iterator_base /* CSH */ {
	 friend class deque<T>;
	 friend class iterator_base;
	 friend class const_iterator;
	 protected:
// CSH	pointer current;
// CSH	pointer first;
// CSH	pointer last;
// CSH	map_pointer node;
	iterator(pointer x, map_pointer y, owner_type* owner) // CSH
		 {	attach(x, y, owner); }
	iterator(owner_type* owner) { attach(0, 0, owner); } // CSH
	 public:
	iterator() /* : current(0), first(0), last(0), node(0) CSH */ {}
// CSH begin
			 iterator(const iterator& b)
			 {  attach(b.current, b.node, b._owner);
			 }

			 iterator& operator=(const iterator& b)
			 {  detach();
				 attach(b.current, b.node, b._owner);
				 return *this;
			 }
			 ~iterator()
			 {  detach();
			 }
// CSH end
	reference operator*() const { assert(valid()); /* CSH */ return *current; }
	difference_type operator-(const iterator& x) const {
		assert(same_owner(x)); // CSH
		 return node == x.node
		? current - x.current
		: difference_type(buffer_size * (node - x.node - 1) +
				  (current - first) + (x.last - x.current));
	}
	iterator& operator++() {
		assert(valid_or_past_end(1)); // CSH
		 if (++current == last) {
		first = *(++node);
		current = first;
		last = first + buffer_size;
		 }
		 return *this;
	}
	iterator operator++(int)  {
		 iterator tmp = *this;
		 ++*this;
		 return tmp;
	}
	iterator& operator--() {
		assert(valid(-1)); // CSH
		 if (current == first) {
		first = *(--node);
		last = first + buffer_size;
		current = last;
		 }
		 --current;
		 return *this;
	}
	iterator operator--(int) {
		 iterator tmp = *this;
		 --*this;
		 return tmp;
	}
	iterator& operator+=(difference_type n) {
		assert(valid_or_past_end(n)); // CSH
		 difference_type offset = n + (current - first);
		 difference_type num_node_to_jump = offset >= 0
		? offset / buffer_size
		: -((-offset + buffer_size - 1) / buffer_size);
		 if (num_node_to_jump == 0)
		current += n;
		 else {
		node = node + num_node_to_jump;
		first = *node;
		last = first + buffer_size;
		current = first + (offset - num_node_to_jump * buffer_size);
		 }
		 return *this;
	}
	iterator& operator-=(difference_type n) { return *this += -n; }
	iterator operator+(difference_type n) const {
		 iterator tmp = *this;
		 return tmp += n;
	}
	iterator operator-(difference_type n) const {
		 iterator tmp = *this;
		 return tmp -= n;
	}
	reference operator[](difference_type n) { return *(*this + n); }
	bool operator==(const iterator& x) const {
		 return current == x.current ||
		((current == first || x.current == x.first) &&
		 *this - x == 0);
	}
	bool operator<(const iterator& x) const {
		assert(same_owner(x)); // CSH
		 return (node == x.node) ? (current < x.current) : (node < x.node);
	}
	 };
	 class const_iterator : public random_access_iterator<T, difference_type>,
		 public iterator_base /* CSH */ {
	 friend class deque<T>;
	 protected:
// CSH	pointer current;
// CSH	pointer first;
// CSH	pointer last;
// CSH	map_pointer node;
	const_iterator(pointer x, map_pointer y, const owner_type* owner) // CSH
		 {	attach(x, y, (/*non-const*/owner_type*)owner); }
	 public:
	const_iterator() /* : current(0), first(0), last(0), node(0) CSH */ {}
	const_iterator(const iterator& x)
// CSH	    : current(x.current), first(x.first), last(x.last), node(x.node) {}
	{ attach(x.current, x.node, x._owner);
	}
// CSH begin
			 const_iterator(const const_iterator& b)
			 {  attach(b.current, b.node, b._owner);
			 }

			 const_iterator& operator=(const const_iterator& b)
			 {  detach();
				 attach(b.current, b.node, b._owner);
				 return *this;
			 }
			 ~const_iterator()
			 {  detach();
			 }
// CSH end
	const_reference operator*() const { assert(valid()); /* CSH */ return *current; }
	difference_type operator-(const const_iterator& x) const {
		 return node == x.node
		? current - x.current
		: difference_type(buffer_size * (node - x.node - 1) +
				  (current - first) + (x.last - x.current));
	}
	const_iterator& operator++() {
		assert(valid_or_past_end(1)); // CSH
		 if (++current == last) {
		first = *(++node);
		current = first;
		last = first + buffer_size;
		 }
		 return *this;
	}
	const_iterator operator++(int)  {
		 const_iterator tmp = *this;
		 ++*this;
		 return tmp;
	}
	const_iterator& operator--() {
		assert(valid(-1)); // CSH
		 if (current == first) {
		first = *(--node);
		last = first + buffer_size;
		current = last;
		 }
		 --current;
		 return *this;
	}
	const_iterator operator--(int) {
		 const_iterator tmp = *this;
		 --*this;
		 return tmp;
	}
	const_iterator& operator+=(difference_type n) {
		assert(valid_or_past_end(n)); // CSH
		 difference_type offset = n + (current - first);
		 difference_type num_node_to_jump = offset >= 0
		? offset / buffer_size
		: -((-offset + buffer_size - 1) / buffer_size);
		 if (num_node_to_jump == 0)
		current += n;
		 else {
		node = node + num_node_to_jump;
		first = *node;
		last = first + buffer_size;
		current = first + (offset - num_node_to_jump * buffer_size);
		 }
		 return *this;
	}
	const_iterator& operator-=(difference_type n) { return *this += -n; }
	const_iterator operator+(difference_type n) const {
		 const_iterator tmp = *this;
		 return tmp += n;
	}
	const_iterator operator-(difference_type n) const {
		 const_iterator tmp = *this;
		 return tmp -= n;
	}
	const_reference operator[](difference_type n) {
		 return *(*this + n);
	}
	bool operator==(const const_iterator& x) const {
		 return current == x.current ||
		((current == first || x.current == x.first) &&
		 *this - x == 0);
	}
	bool operator<(const const_iterator& x) const {
		assert(same_owner(x));
		 return (node == x.node) ? (current < x.current) : (node < x.node);
	}
	 };
	 typedef reverse_iterator<const_iterator, value_type, const_reference,
									  difference_type>  const_reverse_iterator;
	 typedef reverse_iterator<iterator, value_type, reference, difference_type>
		  reverse_iterator;
protected:
	 // CSH iterator start;
	 // CSH iterator finish;
// CSH begin
	map_pointer start_node;
	pointer start_current;
	map_pointer finish_node;
	pointer finish_current;
// CSH end
	 size_type length;
	 map_pointer map;
	 size_type map_size;

	 void allocate_at_begin();
	 void allocate_at_end();
	 void deallocate_at_begin();
	 void deallocate_at_end();

// CSH begin
			 iterator_base* _iter_chain;

			 void orphan_iterators()
			 {  iterator_base* p = _iter_chain;
				 _iter_chain = 0;
				 while (p != 0)
				 {  p->_owner = 0;
					 iterator_base* pnext = p->_next_iter;
					 p->_next_iter = 0;
					 p = pnext;
				 }
			 }

			 void orphan_iterators_at(pointer c, map_pointer n)
			 {  iterator_base* p = _iter_chain;
				 iterator_base* pprev = 0;
				 while (p != 0)
				 {  iterator_base* pnext = p->_next_iter;
					 if (p->current == c && p->node == n)
					 {  p->_owner = 0;
						 p->_next_iter = 0;
						 if (pprev) pprev->_next_iter = pnext;
						 else _iter_chain = pnext;
					 }
					 else
						 pprev = p;
					 p = pnext;
				 }
			 }

			 bool owns(iterator_base p)
			 {	 return p._owner == this;
			 }
// CSH end
public:
	 deque() : /* start(), finish(), CSH */ start_node(0), start_current(0),
	 finish_node(0), finish_current(0), length(0), map(0), map_size(0),
	 _iter_chain(0) /* CSH */ {
	buffer_size = data_allocator.init_page_size();
	 }
	 iterator begin() { return /* start CSH */ iterator(start_current, start_node, this); }
	 const_iterator begin() const { return /* start CSH */ const_iterator(start_current, start_node, this); }
	 iterator end() { return /* finish CSH */ iterator(finish_current, finish_node, this); }
	 const_iterator end() const { return /* finish CSH */ const_iterator(finish_current, finish_node, this); }
	 reverse_iterator rbegin() { return reverse_iterator(end()); }
	 const_reverse_iterator rbegin() const {
		  return const_reverse_iterator(end());
	 }
	 reverse_iterator rend() { return reverse_iterator(begin()); }
	 const_reverse_iterator rend() const {
		  return const_reverse_iterator(begin());
	 }
	 bool empty() const { return length == 0; }
	 size_type size() const { return length; }
	 size_type max_size() const { return data_allocator.max_size(); }
	 reference operator[](size_type n) { return *(begin() + n); }
	 const_reference operator[](size_type n) const { return *(begin() + n); }
	 reference front() { return *begin(); }
	 const_reference front() const { return *begin(); }
	 reference back() { return *(end() - 1); }
	 const_reference back() const { return *(end() - 1); }
	 void push_front(const T& x) {
	 orphan_iterators(); // CSH
	if (empty() || begin().current == begin().first)
		 allocate_at_begin();
	--start_current/* CSH start.current*/;
	construct(start_current/* CSH start.current*/, x);
	++length;
	 }
	 void push_back(const T& x) {
	 orphan_iterators(); // CSH
	if (empty() || end().current == end().last)
		 allocate_at_end();
	construct(finish_current/* CSH finish.current*/, x);
	++finish_current/* CSH finish.current*/;
	++length;
	 }
	 void pop_front() {
	destroy(start_current/* CSH start.current*/);
	orphan_iterators_at(start_current, start_node); // CSH
	++start_current/* CSH start.current*/;
	--length;
	if (empty() || begin().current == begin().last)
		 deallocate_at_begin();
	 }
	 void pop_back() {
	--finish_current/* CSH finish.current*/;
	destroy(finish_current/* CSH finish.current*/);
	orphan_iterators_at(finish_current, finish_node); // CSH
	--length;
	if (empty() || end().current == end().first)
		 deallocate_at_end();
	 }
	 void swap(deque<T>& x) {
// CSH	::swap(start, x.start);
// CSH	::swap(finish, x.finish);
// CSH begin
/* CSH */ std	::swap(start_node, x.start_node);
/* CSH */ std	::swap(start_current, x.start_current);
/* CSH */ std	::swap(finish_node, x.finish_node);
/* CSH */ std	::swap(finish_current, x.finish_current);
// CSH end
/* CSH */ std	::swap(length, x.length);
/* CSH */ std	::swap(map, x.map);
/* CSH */ std	::swap(map_size, x.map_size);
	 }
	 iterator insert(iterator position, const T& x);
	 void insert(iterator position, size_type n, const T& x);
//  template <class Iterator> void insert(iterator position,
//                                        Iterator first, Iterator last);
	 void insert(iterator position, const T* first, const T* last);
	 void erase(iterator position);
	 void erase(iterator first, iterator last);
	 deque(size_type n, const T& value = T())
	: /* start(), finish(), CSH */ start_node(0), start_current(0),
	 finish_node(0), finish_current(0), length(0), map(0), map_size(0), _iter_chain(0) /* CSH */ {
	buffer_size = data_allocator.init_page_size();
	insert(begin(), n, value);
	 }
//  template <class Iterator> deque(Iterator first, Iterator last);
	 deque(const T* first, const T* last)
	: /* start(), finish(), CSH */ start_node(0), start_current(0),
	 finish_node(0), finish_current(0), length(0), map(0), map_size(0), _iter_chain(0) /* CSH */ {
	buffer_size = data_allocator.init_page_size();
	copy(first, last, back_inserter(*this));
	 }
	 deque(const deque<T>& x)
	: /* start(), finish(), CSH */ start_node(0), start_current(0),
	 finish_node(0), finish_current(0), length(0), map(0), map_size(0), _iter_chain(0) /* CSH */ {
	buffer_size = data_allocator.init_page_size();
	copy(x.begin(), x.end(), back_inserter(*this));
	 }
	 deque<T>& operator=(const deque<T>& x) {
	if (this != &x)
		 if (size() >= x.size())
		erase(copy(x.begin(), x.end(), begin()), end());
		 else
// CSH This crashes Borland when deque<T>::iterator gets a destructor
// CSH		copy(x.begin() + size(), x.end(),
//	CSH		  inserter(*this, copy(x.begin(), x.begin() + size(),
//	CSH				  begin())));
// CSH begin
		{  deque<T>::iterator i = copy(x.begin(), x.begin() + size(), begin());
			copy(x.begin() + size(), x.end(), inserter(*this, i));
		}
// CSH end
	return *this;
	 }
	 ~deque();
};

template <class T>
deque<T>::data_allocator_type deque<T>::data_allocator;

template <class T>
deque<T>::map_allocator_type deque<T>::map_allocator;

template <class T>
deque<T>::size_type deque<T>::buffer_size = 0; 
// should be data_allocator.init_page_size(); // Borland bug

template <class T>
bool operator==(const deque<T>& x, const deque<T>& y) {
    return x.size() == y.size() && equal(x.begin(), x.end(), y.begin());
}

template <class T>
bool operator<(const deque<T>& x, const deque<T>& y) {
    return lexicographical_compare(x.begin(), x.end(), y.begin(), y.end());
}

template <class T>
deque<T>::~deque() { while (!empty()) pop_front(); }     

template <class T>
void deque<T>::allocate_at_begin() {
	 pointer p = data_allocator.allocate(buffer_size);
	 if (!empty()) {
	if (start_node/* CSH start.node */ == map) {
		 difference_type i = finish_node/* CSH finish.node */ - start_node/* CSH start.node */;
		 map_size = (i + 1) * 2;
		 map_pointer tmp = map_allocator.allocate(map_size);
		 copy(start_node/* CSH start.node */, finish_node/* CSH finish.node */ + 1, tmp + map_size / 4 + 1);
		 map_allocator.deallocate(map);
		 map = tmp;
		 map[map_size / 4] = p;
		 // CSH start = iterator(p + buffer_size, map + map_size / 4);
		 start_current = p + buffer_size; start_node = map + map_size / 4; // CSH
		 // CSH finish = iterator(finish.current, map + map_size / 4 + i + 1);
		 finish_node = map + map_size / 4 + i + 1; // CSH
	} else {
		 *--start_node/* CSH start.node */ = p;
		 // CSH start = iterator(p + buffer_size, start.node);
		 start_current = p + buffer_size; // CSH
	}
	 } else {
	map_size = map_allocator.init_page_size();
	map = map_allocator.allocate(map_size);
	map[map_size / 2] = p;
	// CSH start = iterator(p + buffer_size / 2 + 1, map + map_size / 2);
	start_current = p + buffer_size / 2 + 1; start_node = map + map_size / 2; // CSH
	// CSH finish = start;
	finish_current = start_current; finish_node = start_node; // CSH
	 }
}

template <class T>
void deque<T>::allocate_at_end() {
	 pointer p = data_allocator.allocate(buffer_size);
	 if (!empty()) {
	if (finish_node/* CSH finish.node */ == map + map_size - 1) {
		 difference_type i = finish_node/* CSH finish.node */ - start_node/* CSH start.node */;
		 map_size = (i + 1) * 2;
		 map_pointer tmp = map_allocator.allocate(map_size);
		 copy(start_node/* CSH start.node */, finish_node/* CSH finish.node */ + 1, tmp + map_size / 4);
		 map_allocator.deallocate(map);
		 map = tmp;
		 map[map_size / 4 + i + 1] = p;
		 // CSH start = iterator(start.current, map + map_size / 4);
		 start_node = map + map_size / 4; // CSH
		 // CSH finish = iterator(p, map + map_size / 4 + i + 1);
		 finish_current = p; finish_node = map + map_size / 4 + i + 1; // CSH.
	} else {
		 *++finish_node/* CSH finish.node */ = p;
		 // CSH finish = iterator(p, finish.node);
		 finish_current = p; // CSH
	}
	 } else {
	map_size = map_allocator.init_page_size();
	map = map_allocator.allocate(map_size);
	map[map_size / 2] = p;
	// CSH start = iterator(p + buffer_size / 2, map + map_size / 2, this);
	start_current = p + buffer_size / 2; start_node = map + map_size / 2; // CSH
	// CSH finish = start;
	finish_current = start_current; finish_node = start_node; // CSH
	 }
}

template <class T>
void deque<T>::deallocate_at_begin() {
	 data_allocator.deallocate(*start_node/* CSH start.node */++);
	 if (empty()) {
	// CSH start = iterator();
	start_current = 0; start_node = 0; // CSH
	// CSH finish = start;
	finish_current = 0; finish_node = 0; // CSH
	map_allocator.deallocate(map);
	 } else
	// CSH start = iterator(*start.node, start.node);
	start_current = *start_node; // CSH
}

template <class T>
void deque<T>::deallocate_at_end() {
	 data_allocator.deallocate(*finish_node/* CSH finish.node */--);
	 if (empty()) {
	// CSH start = iterator();
	start_current = 0; start_node = 0; // CSH
	// CSH finish = start;
	finish_current = 0; finish_node = 0; // CSH
	map_allocator.deallocate(map);
	 } else
	// finish = iterator(*finish.node + buffer_size, finish.node);
	finish_current = *finish_node + buffer_size; // CSH
}

template <class T>
deque<T>::iterator deque<T>::insert(iterator position, const T& x) {
	 if (position == begin()) {
	push_front(x);
	return begin();
	 } else if (position == end()) {
	push_back(x);
	return end() - 1;
	 } else {
	difference_type index = position - begin();
	if (index < length) {
		 push_front(*begin());
		 copy(begin() + 2, begin() + index + 1, begin() + 1);
	} else {
		 push_back(*(end() - 1));
		 copy_backward(begin() + index, end() - 2, end() - 1);
		  }
	*(begin() + index) = x;
	return begin() + index;
	 }
}

template <class T>
void deque<T>::insert(iterator position, size_type n, const T& x) {
	assert(owns(position)); // CSH
	 difference_type index = position - begin();
	 difference_type remainder = length - index;
	 if (remainder > index) {
	if (n > index) {
		 difference_type m = n - index;
		 while (m-- > 0) push_front(x);
		 difference_type i = index;
		 while (i--) push_front(*(begin() + n - 1));
		 fill(begin() + n, begin() + n + index, x);
	} else {
		 difference_type i = n;
		 while (i--) push_front(*(begin() + n - 1));
		 copy(begin() + n + n, begin() + n + index, begin() + n);
		 fill(begin() + index, begin() + n + index, x);
	}
	 } else {
	difference_type orig_len = index + remainder;
	if (n > remainder) {
		 difference_type m = n - remainder;
		 while (m-- > 0) push_back(x);
		 difference_type i = 0;
		 while (i < remainder) push_back(*(begin() + index + i++));
		 fill(begin() + index, begin() + orig_len, x);
	} else {
		 difference_type i = 0;
		 while (i < n) push_back(*(begin() + orig_len - n + i++));
		 copy_backward(begin() + index, begin() + orig_len - n,
			  begin() + orig_len);
		 fill(begin() + index, begin() + index + n, x);
	}
	 }
}

template <class T>
void deque<T>::insert(iterator position, const T* first, const T* last) {
	assert(owns(position)); // CSH
	 difference_type index = position - begin();
	 difference_type remainder = length - index;
	 size_type n = 0;
	 distance(first, last, n);
	 if (remainder > index) {
	if (n > index) {
		 const T* m = last - index;
		 while (m != first) push_front(*--m);
		 difference_type i = index;
		 while (i--) push_front(*(begin() + n - 1));
		 copy(last - index, last, begin() + n);
	} else {
		 difference_type i = n;
		 while (i--) push_front(*(begin() + n - 1));
		 copy(begin() + n + n, begin() + n + index, begin() + n);
		 copy(first, last, begin() + index);
	}
	 } else {
	difference_type orig_len = index + remainder;
	if (n > remainder) {
		 const T* m = first + remainder;
		 while (m != last) push_back(*m++);
		 difference_type i = 0;
		 while (i < remainder) push_back(*(begin() + index + i++));
		 copy(first, first + remainder, begin() + index);
	} else {
		 difference_type i = 0;
		 while (i < n) push_back(*(begin() + orig_len - n + i++));
		 copy_backward(begin() + index, begin() + orig_len - n,
			  begin() + orig_len);
		 copy(first, last, begin() + index);
	}
	 }
}

template <class T>
void deque<T>::erase(iterator position) {
	assert(owns(position)); // CSH
	 if (end() - position > position - begin()) {
	copy_backward(begin(), position, position + 1);
	pop_front();
	 } else {
	copy(position + 1, end(), position);
	pop_back();
	 }
}

template <class T>
void deque<T>::erase(iterator first, iterator last) {
	assert(owns(first) && owns(last)); // CSH
	 difference_type n = last - first;
	 if (end() - last > first - begin()) {
	copy_backward(begin(), first, last);
	while(n-- > 0) pop_front();
	 } else   {
	copy(last, end(), first);
	while(n-- > 0) pop_back();
	 }
}


#undef Allocator
#undef deque

// CSH begin
} // namespace std
// CSH end

#endif
